import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import peasy.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class orbital extends PApplet {

/*
Author : Eric Pavey
 Date : 2013-11-17
 
 4 Weeks of creativity, day 7
 Final day of working on this sketch.
 
 Overiew:
 Lights orbit a central star, smashing into particles.  If the particles
 are knocked outside a bounding box, they are pulled back in.  Otherwise
 they come to rest.  Will be in a window just slightly smaller than the screen.
 

 Learning today:  Access to OpenGl : Doesn't seem to work in Processing 2.0 
 on Mac 10.8.5 :(  I had hoped to replace the lights with glowing sprites,
 but since I can't seem to access jogl properly, no dice.  Fixed bug in
 tail drawing code that wasn't changing the strokeWeight based on camera
 distance.
 
 Mouse tumbles camera around balls. 's' saves an image.  r saves a range of 
 frames for movie making.
 
 2d space (mouse query):  
 +X is left to right
 +Y is top to bottom
 3d space:
 +X is left to right 
 +Y is top to bottom
 +Z is back to front
 */

// NEED PEASYCAM INSTALLED!
// http://mrfeinberg.com/peasycam/
// http://mrfeinberg.com/peasycam/reference/index.html


//------------------------------------------------------------------------
PeasyCam cam;

int maxBalls = 256;
// Max 3:
int maxLights = 2;  
// Defines the size of the oveall environment.
float maxSize = 512;
int ballIndex = 0;

// used to lod out spheres as they move away from the camera
int gMinBallRes = 8;
int gMaxBallRes = 16; // default sphere detail

// The center of the world:
PVector gCenter = new PVector(0, 0, 0);
// Global gravity used for attraction to center of world:
float gGravity = 2;

ArrayList balls = new ArrayList();
ArrayList lights = new ArrayList();
float gStrokeMult;

boolean recording = false;

//------------------------------------------------------------------------

public void setup() {
  size(displayWidth-4, displayHeight-64, P3D);
  // For movie recording:
  //size(640, 480, P3D);
  
  // magic number to fix the size of strokes at different screen resolutions.
  gStrokeMult = (width+height)/50;

  cam = new PeasyCam(this, maxSize);
  cam.setMinimumDistance(32);
  cam.setMaximumDistance(maxSize*4);
  smooth();

  // for drawing the light trails.
  strokeJoin(ROUND);
  strokeCap(ROUND);

  // Light creation:
  for (int i=0; i<maxLights; i++) {
    Light light = new Light(i);
    lights.add(light);
  }

  background(16);
  frame.setTitle("Orbital");
  frameRate(30);
}

//------------------------------------------------------------------------

public void draw() {
  background(16);

  fill(255, 255, 255);
  sphere(maxSize/64);

  // Lighting:
  for (int i=0; i<maxLights; i++) {
    Light light = (Light) lights.get(i);
    light.compute();
  }
  directionalLight(128, 64, 0, 0, 1, 0);
  pointLight(255, 255, 255, 0, 0, 0);

  // create the bounding box:
  noFill();
  stroke(32);
  strokeWeight(2);
  box(maxSize);

  // ball creation:
  if (balls.size() < maxBalls) {
    Ball c = new Ball();
    balls.add(ballIndex, c);
    ballIndex++;
  }

  // Ball Rendering:
  for (int i=0; i<balls.size(); i++) {
    Ball ballr = (Ball) balls.get(i);

    // test for light collision:
    for (int j=0; j<maxLights; j++) {
      Light light = (Light) lights.get(j);
      ballr.collideLight(light);
    }

    // test for ball collisions:
    for (int j=0; j<balls.size(); j++) {
      if (i==j) {
        continue;
      }
      Ball otherBall = (Ball) balls.get(j);
      ballr.collideBall(otherBall);
    }

    ballr.update();
    ballr.render();
  }
  if (recording) {
    saveFrame("output/frames####.png");
  }
}


//------------------------------------------------------------------------
// utils

public void keyPressed()
{
  // Note the single quotes ('), not double quotes(")
  if (key == 's') {
    saveFrame("4woc07_####.jpg");
  }
  if (key == 'r' || key == 'R') {
    recording = !recording;
    println("Recording: " + recording);
  }  
}

public float percentBetween(float val, float origMin, float origMax) {
  if (val <= origMin) {
    return 0;
  }
  else if (val >= origMax) {
    return 1;
  }
  else {
    return (val - origMin) / (origMax - origMin);
  }
}

class Ball { 
  /*
  Fantastic ball class.
   */
  // Track current position:
  PVector pos;
  // Track current direction & velocity;
  PVector dir;
  // size of the ball
  float bsize;
  // How much to damp the physics.  Higher numbers = less damping.
  float damp = .95f;
  // Mass of ball, based on its size.
  float mass;

  Ball() {  
    pos = new PVector(random(-maxSize/2, maxSize/2), random(-maxSize/2, maxSize/2), random(-maxSize/2, maxSize/2));
    dir = new PVector(random(-1, 1), random(-1, 1), random(-1, 1));
    bsize = random(8, maxSize/16);
    mass = PI * pow(bsize, 3);
    //bsize = 64;
  } 

  public void update() {
    // Compute the current position.  collide() should be called before this.

    // If ball is outside max size, the bring it back.
    float distToCenter = dist(pos.x, pos.y, pos.z, 0, 0, 0); 
    if (distToCenter > maxSize/2) {
      // The further away the ball is, the greater the force on it to bring it back:
      float percent = percentBetween(distToCenter, maxSize/2, maxSize*2);
      float force = percent * gGravity;

      PVector vecToCenter = PVector.sub(gCenter, pos);
      vecToCenter.normalize();
      vecToCenter.mult(force);
      dir.add(vecToCenter);
    }

    dir.mult(damp);
    pos = PVector.add(pos, dir);
  }

  public void render() {
    // Draw our ball
    
    // LOD out ball based on distance to camera:
    float[] camPos = cam.getPosition();
    float distToCam = dist(pos.x, pos.y, pos.z, camPos[0], camPos[1], camPos[2]);
    float pct = percentBetween(distToCam, maxSize/4, maxSize*4);
    int detail = PApplet.parseInt(lerp(gMaxBallRes, gMinBallRes, pct));
    
    noStroke();
    fill(128);
    pushMatrix();
    translate(pos.x, pos.y, pos.z);
    sphereDetail(detail);
    sphere(bsize);
    popMatrix();
  }

  public void collideLight(Light light) {
    // Test to see if this ball has collided with a light.
    // get the distance between the ball and the light:
    float d = dist(pos.x, pos.y, pos.z, light.pos.x, light.pos.y, light.pos.z); 
    // Sum the size of the light and ball:
    float sizeAdd = bsize + light.lsize*1.25f;
    // if that size is greater than the distance between, collision!
    if (d < sizeAdd) {
      // get the vector from light to ball, which is the force at point of collision:
      PVector lightToBall = PVector.sub(pos, light.pos);
      lightToBall.normalize();
      lightToBall.mult(light.dir.mag());
      dir.add(lightToBall);
    }
  }

  public void collideBall(Ball otherBall) {
    // Test to see if this ball has collided with another ball.
    // Get the distance between the ball and the otherBall:
    float d = dist(pos.x, pos.y, pos.z, otherBall.pos.x, otherBall.pos.y, otherBall.pos.z); 
    // Sum the size of the light and ball:
    float sizeAdd = bsize + otherBall.bsize;
    // if that size is greater than the distance between, collision!
    if (d < sizeAdd) {
      // Get the vector from each ball to each other, since they will both influence
      // one another.
      PVector thisToOther = PVector.sub(otherBall.pos, pos);
      PVector otherToThis = PVector.sub(pos, otherBall.pos);
      thisToOther.normalize();
      otherToThis.normalize();

      // Figure out how much each ball influences each other based on mass:
      float thisForceMult = 0;
      float otherForceMult = 0;
      if (otherBall.mass <= mass) {
        otherForceMult = otherBall.mass / mass;
        thisForceMult = 1 - otherForceMult;
      }
      else {
        thisForceMult = mass / otherBall.mass;
        otherForceMult = 1 - thisForceMult;
      }


      thisToOther.mult(dir.mag() * thisForceMult);
      otherToThis.mult(otherBall.dir.mag() * otherForceMult);

      dir.add(otherToThis);
      otherBall.dir.add(thisToOther);
    }
  }
} 

class Light {
  /*
  Our mobile light source.
   */
  // predefined colors that can be used if more than one light is
  // created.  Only allows for 3 lights max, unless you add more colors.
  int colors[] = {
    color(64, 0, 255), color(0, 64, 255), color(64, 64, 255)
  };
  int[] col;
  // number of sphere's to draw in the trail:
  int lightTrail = 64;
  ArrayList trails = new ArrayList(); 
  // Size of this light:
  float lsize = maxSize/32;

  // Help to give our light an initial 'push' before it starts orbiting:
  float dirForce = 20;
  PVector pos = new PVector(random(-maxSize/2, maxSize/2), random(-maxSize/2, maxSize/2), random(-maxSize/2, maxSize/2));
  PVector dir = new PVector(random(-dirForce, dirForce), random(-dirForce, dirForce), random(-dirForce, dirForce));

  Light(int colorIndex) {
    int c = colors[colorIndex];
    col = new int[] {
      PApplet.parseInt(red(c)), PApplet.parseInt(green(c)), PApplet.parseInt(blue(c))
    };
    for (int i=0; i<lightTrail-1; i++) {
      trails.add(pos.get());
    }
  }

  public void compute() {
    update();
    render();
  }

  public void update() {
    PVector vecToCenter = PVector.sub(gCenter, pos);
    vecToCenter.normalize();
    vecToCenter.mult(gGravity);
    dir.add(vecToCenter);
    pos.add(dir);

    // fill our buffer of light trails:
    trails.add(pos.get());
    trails.remove(0);
  }

  public void render() {

    // Draw the trail:
    noFill();
    stroke(col[0], col[1], col[2]);
    beginShape();
    for (int i=0; i<lightTrail-1; i++) {
      // Scale the tail based on its distane from the light
      float sVal = PApplet.parseFloat(i)/PApplet.parseFloat(lightTrail);
      
      PVector trailPos = (PVector) trails.get(i);
      // Stroke weights are a fixed size on screen, regardless of
      // zoom.  So we need to fake the zoom effect.
      float[] camPos = cam.getPosition();
      float distToCam = dist(trailPos.x, trailPos.y, trailPos.z, camPos[0], camPos[1], camPos[2]);
      float mult = 1/(distToCam / (lsize*2)) * gStrokeMult;
      strokeWeight(sVal*lsize*mult);
      curveVertex(trailPos.x, trailPos.y, trailPos.z);
    }
    curveVertex(pos.x, pos.y, pos.z);
    endShape();
    noStroke();

    // Draw the main light:
    sphereDetail(16);
    pushMatrix();
    translate(pos.x, pos.y, pos.z);
    fill(col[0], col[1], col[2]);
    
    sphere(lsize);
    pointLight(col[0], col[1], col[2], 0, 0, 0);
    
    //pointLight(col[0], col[1], col[2], 0, 0, 0);
    popMatrix();
  }
}

  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "orbital" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
