/*
Author : Eric Pavey
 Date : 2013-11-17
 
 4 Weeks of creativity, day 7
 Final day of working on this sketch.
 
 Overiew:
 Lights orbit a central star, smashing into particles.  If the particles
 are knocked outside a bounding box, they are pulled back in.  Otherwise
 they come to rest.  Will be in a window just slightly smaller than the screen.
 

 Learning today:  Access to OpenGl : Doesn't seem to work in Processing 2.0 
 on Mac 10.8.5 :(  I had hoped to replace the lights with glowing sprites,
 but since I can't seem to access jogl properly, no dice.  Fixed bug in
 tail drawing code that wasn't changing the strokeWeight based on camera
 distance.
 
 Mouse tumbles camera around balls. 's' saves an image.  r saves a range of 
 frames for movie making.
 
 2d space (mouse query):  
 +X is left to right
 +Y is top to bottom
 3d space:
 +X is left to right 
 +Y is top to bottom
 +Z is back to front
 */

// NEED PEASYCAM INSTALLED!
// http://mrfeinberg.com/peasycam/
// http://mrfeinberg.com/peasycam/reference/index.html
import peasy.*;

//------------------------------------------------------------------------
PeasyCam cam;

int maxBalls = 256;
// Max 3:
int maxLights = 2;  
// Defines the size of the oveall environment.
float maxSize = 512;
int ballIndex = 0;

// used to lod out spheres as they move away from the camera
int gMinBallRes = 8;
int gMaxBallRes = 16; // default sphere detail

// The center of the world:
PVector gCenter = new PVector(0, 0, 0);
// Global gravity used for attraction to center of world:
float gGravity = 2;

ArrayList balls = new ArrayList();
ArrayList lights = new ArrayList();
float gStrokeMult;

boolean recording = false;

//------------------------------------------------------------------------

void setup() {
  size(displayWidth-4, displayHeight-64, P3D);
  // For movie recording:
  //size(640, 480, P3D);
  
  // magic number to fix the size of strokes at different screen resolutions.
  gStrokeMult = (width+height)/50;

  cam = new PeasyCam(this, maxSize);
  cam.setMinimumDistance(32);
  cam.setMaximumDistance(maxSize*4);
  smooth();

  // for drawing the light trails.
  strokeJoin(ROUND);
  strokeCap(ROUND);

  // Light creation:
  for (int i=0; i<maxLights; i++) {
    Light light = new Light(i);
    lights.add(light);
  }

  background(16);
  frame.setTitle("Orbital");
  frameRate(30);
}

//------------------------------------------------------------------------

void draw() {
  background(16);

  fill(255, 255, 255);
  sphere(maxSize/64);

  // Lighting:
  for (int i=0; i<maxLights; i++) {
    Light light = (Light) lights.get(i);
    light.compute();
  }
  directionalLight(128, 64, 0, 0, 1, 0);
  pointLight(255, 255, 255, 0, 0, 0);

  // create the bounding box:
  noFill();
  stroke(32);
  strokeWeight(2);
  box(maxSize);

  // ball creation:
  if (balls.size() < maxBalls) {
    Ball c = new Ball();
    balls.add(ballIndex, c);
    ballIndex++;
  }

  // Ball Rendering:
  for (int i=0; i<balls.size(); i++) {
    Ball ballr = (Ball) balls.get(i);

    // test for light collision:
    for (int j=0; j<maxLights; j++) {
      Light light = (Light) lights.get(j);
      ballr.collideLight(light);
    }

    // test for ball collisions:
    for (int j=0; j<balls.size(); j++) {
      if (i==j) {
        continue;
      }
      Ball otherBall = (Ball) balls.get(j);
      ballr.collideBall(otherBall);
    }

    ballr.update();
    ballr.render();
  }
  if (recording) {
    saveFrame("output/frames####.png");
  }
}


//------------------------------------------------------------------------
// utils

void keyPressed()
{
  // Note the single quotes ('), not double quotes(")
  if (key == 's') {
    saveFrame("4woc07_####.jpg");
  }
  if (key == 'r' || key == 'R') {
    recording = !recording;
    println("Recording: " + recording);
  }  
}

float percentBetween(float val, float origMin, float origMax) {
  if (val <= origMin) {
    return 0;
  }
  else if (val >= origMax) {
    return 1;
  }
  else {
    return (val - origMin) / (origMax - origMin);
  }
}

